using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using MacGruber;
using UnityEngine;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.Serialization;


namespace CheesyFX {

	public class ReadMyLips : MVRScript
	{
		public static ReadMyLips singleton;
		private GenerateDAZMorphsControlUI morphControl;
		private static System.Random rng = new System.Random();  
		// private static Dictionary<JSONStorableFloat, float> painMorphps= new Dictionary<JSONStorableFloat, float>();
		// private static Dictionary<JSONStorableFloat, float> joyMorphs= new Dictionary<JSONStorableFloat, float>();
		// private static Dictionary<JSONStorableFloat, float> lastMorphs= new Dictionary<JSONStorableFloat, float>();
		private static List<Morph> painMorphs = new List<Morph>();
		private static List<Morph> joyMorphs = new List<Morph>();
		private static List<Morph> lastMorphs = new List<Morph>();
		public static JSONStorableFloat maxMorphVal = CheesyUtils.SetupJSONFloat("Max Morph Value", .1f, 0f, 1f);
		public static JSONStorableFloat painTreshold = CheesyUtils.SetupJSONFloat("Pain Thresold", .7f, 0f, 1f, true);
		public static JSONStorableFloat joyExpressionScale = CheesyUtils.SetupJSONFloat("Joy Expression Scale", 20f, 1f, 40f, true);
		public static JSONStorableFloat painExpressionScale = CheesyUtils.SetupJSONFloat("Pain Expression Scale", 10f, 1f, 20f, true);
		public static JSONStorableFloat gloabalSensitivity = CheesyUtils.SetupJSONFloat("Global Sensitivity", 1f, 0f, 5f);
		
		public static JSONStorableFloat stimulation = new JSONStorableFloat("Stimulation", 0f, 0f, 1f, true, false);
		public static bool isReset = false;
		private static float resetTargetsTimer;
		private static float stimBurstTimer;
		private static bool _slapped;
		private static bool slapped{
			get{return _slapped;}
			set{
				_slapped = value;
				if(!value){
					foreach(var morph in painMorphs){
						morph.target = 0f;
					}
					foreach(var morph in joyMorphs){
						morph.target = 0f;
					}
				}
			}
		}
		public static DAZMeshEyelidControl eyelidBehavior;
		public static float blinkTimeout;

		public override void Init()
		{
			singleton = this;
			eyelidBehavior = (DAZMeshEyelidControl) SlapMe.atom.GetStorableByID("EyelidControl");
			DAZCharacterSelector characterSelector = SlapMe.atom.GetStorableByID("geometry") as DAZCharacterSelector;
			morphControl = characterSelector.morphsControlUI;
			string joyMorphPath = SlapMe.packageUid + "Custom/Atom/Person/Morphs/female/CheesyFX/BodyLanguage/Expressions/Joy/";
			string painMorphPath = SlapMe.packageUid + "Custom/Atom/Person/Morphs/female/CheesyFX/BodyLanguage/Expressions/Displease";
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "Ascorad"))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "AshAuryn"))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(painMorphPath))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				painMorphs.Add(GetMorph(pathNormalized));
			}
			joyMorphs.Add(GetMorph("Desire"));
			joyMorphs.Add(GetMorph("Excitement"));
			joyMorphs.Add(GetMorph("Flirting"));
			joyMorphs.Add(GetMorph("Flirting Feminine Left"));
			joyMorphs.Add(GetMorph("Flirting Feminine Right"));
			joyMorphs.Add(GetMorph("Flirting Masculine Left"));
			joyMorphs.Add(GetMorph("Flirting Masculine Right"));
			joyMorphs.Add(GetMorph("Frown"));
			joyMorphs.Add(GetMorph("Sad"));
			joyMorphs.Add(GetMorph("Scream"));
			joyMorphs.Add(GetMorph("Shock"));
			joyMorphs.Add(GetMorph("Smile Full Face"));
			joyMorphs.Add(GetMorph("Snarl Left"));
			joyMorphs.Add(GetMorph("Snarl Right"));
			joyMorphs.Add(GetMorph("Surprise"));
			
			// joyMorphpsBrows.Add(GetMorph("Brow Down"));
			joyMorphs.Add(GetMorph("Brow Down Left"));
			joyMorphs.Add(GetMorph("Brow Down Right"));
			// joyMorphps.Add(GetMorph("Brow Inner Down"));
			joyMorphs.Add(GetMorph("Brow Inner Down Left"));
			joyMorphs.Add(GetMorph("Brow Inner Down Right"));
			joyMorphs.Add(GetMorph("Brow Inner Up Left"));
			joyMorphs.Add(GetMorph("Brow Inner Up Right"));
			// joyMorphps.Add(GetMorph("Brow Outer Down Left"));
			joyMorphs.Add(GetMorph("Brow Outer Down Left"));
			joyMorphs.Add(GetMorph("Brow Outer Down Right"));
			// joyMorphps.Add(GetMorph("Brow Outer Up"));
			joyMorphs.Add(GetMorph("Brow Outer Up Left"));
			joyMorphs.Add(GetMorph("Brow Outer Up Right"));
			// joyMorphps.Add(GetMorph("Brow Up"));
			joyMorphs.Add(GetMorph("Brow Up Left"));
			joyMorphs.Add(GetMorph("Brow Up Right"));
			joyMorphs.Add(GetMorph("Brow Squeeze"));
			
			joyMorphs.Add(GetMorph("Eyes Closed"));
			joyMorphs.Add(GetMorph("Eyes Squint"));
			joyMorphs.Add(GetMorph("Eyes Squint Left"));
			joyMorphs.Add(GetMorph("Eyes Squint Right"));
			
			// joyMorphs.Add(GetMorph("Mouth Open"));
			// joyMorphs.Add(GetMorph("Mouth Open Wide"));
			// joyMorphs.Add(GetMorph("Mouth Narrow"));
			// joyMorphs.Add(GetMorph("Mouth Narrow Left"));
			// joyMorphs.Add(GetMorph("Mouth Narrow Right"));
			joyMorphs.Add(GetMorph("Mouth Side-Side Left"));
			joyMorphs.Add(GetMorph("Mouth Side-Side Right"));
			joyMorphs.Add(GetMorph("Mouth Smile"));
			joyMorphs.Add(GetMorph("Mouth Smile Open"));
			// joyMorphs.Add(GetMorph("Mouth Smile Simple"));
			joyMorphs.Add(GetMorph("Mouth Smile Simple Left"));
			joyMorphs.Add(GetMorph("Mouth Smile Simple Right"));
			
			joyMorphs.Add(GetMorph("Lip Bottom Down"));
			joyMorphs.Add(GetMorph("Lip Bottom Out"));
			// joyMorphps.Add(GetMorph("Lip Bottom In"));
			joyMorphs.Add(GetMorph("Lip Bottom In Left"));
			joyMorphs.Add(GetMorph("Lip Bottom Out Right"));
			joyMorphs.Add(GetMorph("Lip Bottom Up"));
			joyMorphs.Add(GetMorph("Lip Bottom Up Left"));
			joyMorphs.Add(GetMorph("Lip Bottom Up Right"));
			// joyMorphs.Add(GetMorph("Lip Top Up"));
			joyMorphs.Add(GetMorph("Lip Top Up Left"));
			joyMorphs.Add(GetMorph("Lip Top Up Right"));
			joyMorphs.Add(GetMorph("Lips Close"));
			joyMorphs.Add(GetMorph("Lips Part"));
			// joyMorphs.Add(GetMorph("Lips Pucker"));
			joyMorphs.Add(GetMorph("Lips Pucker Wide"));
			enabledJSON.setCallbackFunction += val => Reset();
			CreateUI();
			Reset();
		}

		private void CreateUI()
		{
			stimulation.CreateUI(this);
			painTreshold.CreateUI(this, true);
			maxMorphVal.CreateUI(this);
			joyExpressionScale.CreateUI(this);
			painExpressionScale.CreateUI(this);
			maxMorphVal.CreateUI(this);
			stimulation.CreateUI(this);
		}

		public static void Stimulate(float stimulus){
			float stim = stimulus * gloabalSensitivity.val*.01f;
			stimulation.val += stim;
			isReset = false;
			slapped = true;
			stimBurstTimer = .2f;
			DistributeStimulus(stim);
		}

		public static void DistributeStimulus(float stim){
			float available = stim;
			List<Morph> morphs;
			float factor;
			if(stimulation.val > painTreshold.val){
				morphs = new List<Morph>(painMorphs);
				factor = painExpressionScale.val;
				// factor = Mathf.Lerp(painExpressionScale.val, painExpressionScale.val * .05f,
				// 	(stimulation.val - painTreshold.val)/(1f - painTreshold.val));
			}
			else{
				morphs = new List<Morph>(joyMorphs);
				factor = joyExpressionScale.val;
				factor = Mathf.Lerp(joyExpressionScale.val, joyExpressionScale.val * .25f, stimulation.val);
			}

			
			available *= factor;
			List<Morph> rndMorphs;
			if (lastMorphs.Count > 0 && rng.Next(10) == 0)
			{
				rndMorphs = new List<Morph>(lastMorphs);
				lastMorphs.Clear();
			}
			else
			{
				rndMorphs = morphs.OrderBy(x => rng.Next()).ToList();
				// lastMorphs = lastMorphs.Skip((int)(lastMorphs.Count * .5f));
				rndMorphs.InsertRange(0, lastMorphs.Skip(Mathf.CeilToInt(lastMorphs.Count * .75f)));
				lastMorphs.Clear();
			}
			foreach(var morph in rndMorphs){
				// 
				float increment = UnityEngine.Random.Range(0f, Math.Min(available, Mathf.Abs(maxMorphVal.val - morph.val)));
				morph.target += increment;
				available -= increment;
				lastMorphs.Add(morph);
				// increment.Print();
				if (available <= .05f) break;
			}
			targetsReset = false;
			// "popo".Print();
			// lastMorphs.Count.Print();
		}

		private static bool targetsReset;
		public void Update(){
			if(enabled)
			{
				if(blinkTimeout > 0f) blinkTimeout -= Time.deltaTime;
				if(stimulation.val > .01f){
					stimulation.val = Mathf.Lerp(stimulation.val, 0f, Time.deltaTime);
				}
				else if(!isReset)
				{
					ResetTargets();
				}
				// if(timer > 0f){
					float deltaT = 4f*Time.deltaTime;;
					if(resetTargetsTimer > 0f) resetTargetsTimer -= Time.deltaTime;
					if(stimBurstTimer > 0f) stimBurstTimer -= Time.deltaTime;
					if(resetTargetsTimer<=0f) slapped = false;
					foreach(var morph in painMorphs)
					{
						morph.LerpToTarget();
					}
					foreach(var morph in joyMorphs){
						morph.LerpToTarget();
					}
				// }
				if(!targetsReset && (resetTargetsTimer <= 0f || stimBurstTimer <= 0f))
				{
					// stimBurstTimer.Print();
					// "timer".Print();
					// if (isReset) return;
					// isReset.Print();
					foreach(var morph in painMorphs){
						if(morph.target > 0f) morph.target = 0f;
					}
					foreach(var morph in joyMorphs){
						// morph.val = Mathf.Lerp(morph.val, 0f, 2f*Time.deltaTime);
						if(morph.target > 0f) morph.target = 0f;
					}

					targetsReset = true;
					resetTargetsTimer = 1f;
				}
			}
			
		}

		// void AdjustNipples(float erection){
		// 	nippleSpring.val = Mathf.Lerp(nippleSpring.defaultVal, 15f, erection);
		// 	nippleDamper.val = Mathf.Lerp(nippleDamper.defaultVal, 3f, erection);
		// 	nippleMorph.jsonFloat.val = erection;
		// }

		private void ResetTargets()
		{
			foreach(var morph in painMorphs)
			{
				morph.target = 0f;
			}
			
			foreach (var morph in joyMorphs)
			{
				morph.target = 0f;
			}
			isReset = true;
		}

		public static void Reset(){
			foreach(var morph in painMorphs)
			{
				morph.val = 0f;
				morph.target = 0f;
			}
			
			foreach (var morph in joyMorphs)
			{
				morph.val = 0f;
				morph.target = 0f;
			}
			stimulation.val = 0f;
			isReset = true;
		}

		private Morph GetMorph(string uid)
		{
			DAZMorph morph = morphControl.GetMorphByUid(uid);
			if (morph == null) uid.Print();
			return new Morph(morph);
		}

		public class Morph
		{
			private DAZMorph dazMorph;
			public string name => dazMorph.displayName;
			// private JSONStorableFloat jsonfloat;
			public float target;

			public float val
			{
				get { return dazMorph.morphValue; }
				set { dazMorph.morphValue = value; }
			}

			public Morph(DAZMorph dazMorph)
			{
				this.dazMorph = dazMorph;
			}

			public void LerpToTarget()
			{
				if(Mathf.Abs(val - target) < .001f) return;
				if(target > .001f) val = Mathf.Lerp(val, target, 4f*Time.deltaTime);
				else val = Mathf.Lerp(val, target, 2f*Time.deltaTime);
			}
		}
	}
}